/*
 * ToRegExpAlgebraic.cpp
 *
 *  Created on: 11. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "ToRegExpAlgebraic.h"

#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/EpsilonNFA.h>

#include <equations/RightRegularEquationSolver.h>
#include <regexp/unbounded/UnboundedRegExpElements.h>
#include <registration/AlgoRegistration.hpp>

namespace automaton {

namespace convert {

regexp::UnboundedRegExp < > ToRegExpAlgebraic::convert( const automaton::EpsilonNFA < > & automaton ) {
	equations::RightRegularEquationSolver < DefaultSymbolType > solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( DefaultSymbolType( q ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( DefaultSymbolType( q ), regexp::UnboundedRegExpEpsilon < DefaultSymbolType > { } );
	}

	for( const auto & p : automaton.getSymbolTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( DefaultSymbolType( p.first.first ), DefaultSymbolType( q ), regexp::UnboundedRegExpSymbol < DefaultSymbolType > { p.first.second } );
		}
	}

	for( const auto & p : automaton.getEpsilonTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( DefaultSymbolType( p.first ), DefaultSymbolType( q ), regexp::UnboundedRegExpEpsilon < DefaultSymbolType > { } );
		}
	}

	return solver.solve( DefaultSymbolType( automaton.getInitialState() ) );
}

regexp::UnboundedRegExp < > ToRegExpAlgebraic::convert( const automaton::MultiInitialStateNFA < > & automaton ) {
	equations::RightRegularEquationSolver < DefaultSymbolType > solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( DefaultSymbolType( q ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( DefaultSymbolType( q ), regexp::UnboundedRegExpEpsilon < DefaultSymbolType > { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( DefaultSymbolType( p.first.first ), DefaultSymbolType( q ), regexp::UnboundedRegExpSymbol < DefaultSymbolType > { p.first.second } );
		}
	}

	regexp::UnboundedRegExpAlternation < DefaultSymbolType > alternation;
	for(const auto& initialSymbol : automaton.getInitialStates() ) {
		// set symbol for which the solver will solve equation system
		DefaultSymbolType tmp { initialSymbol };

		alternation.appendElement( solver.solve( tmp ).getRegExp().getStructure() );
	}

	return regexp::UnboundedRegExp < > { regexp::UnboundedRegExpStructure < DefaultSymbolType > ( alternation ) };
}

regexp::UnboundedRegExp < > ToRegExpAlgebraic::convert( const automaton::NFA < > & automaton ) {
	equations::RightRegularEquationSolver < DefaultSymbolType > solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( DefaultSymbolType( q ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( DefaultSymbolType( q ), regexp::UnboundedRegExpEpsilon < DefaultSymbolType > { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		for( const auto & q : p.second ) {
			solver.addEquation( DefaultSymbolType( p.first.first ), DefaultSymbolType( q ), regexp::UnboundedRegExpSymbol < DefaultSymbolType > { p.first.second } );
		}
	}

	return solver.solve( DefaultSymbolType( automaton.getInitialState() ) );
}

regexp::UnboundedRegExp < > ToRegExpAlgebraic::convert( const automaton::DFA < > & automaton ) {
	equations::RightRegularEquationSolver < DefaultSymbolType > solver;

	// initialize equations
	for( const auto & q : automaton.getStates( ) )
		solver.addSymbol( DefaultSymbolType( q ) );

	for( const auto & q : automaton.getStates( ) ) {
		if( automaton.getFinalStates( ).count( q ) > 0 )
			solver.addEquation( DefaultSymbolType( q ), regexp::UnboundedRegExpEpsilon < DefaultSymbolType > { } );
	}

	for( const auto & p : automaton.getTransitions() ) {
		solver.addEquation( DefaultSymbolType( p.first.first ), DefaultSymbolType( p.second ), regexp::UnboundedRegExpSymbol < DefaultSymbolType > { p.first.second } );
	}

	return solver.solve( DefaultSymbolType( automaton.getInitialState() ) );
}

auto ToRegExpAlgebraicEpsilonNFA = registration::AbstractRegister < ToRegExpAlgebraic, regexp::UnboundedRegExp < >, const automaton::EpsilonNFA < > & > ( ToRegExpAlgebraic::convert );
auto ToRegExpAlgebraicMultiInitialStateNFA = registration::AbstractRegister < ToRegExpAlgebraic, regexp::UnboundedRegExp < >, const automaton::MultiInitialStateNFA < > & > ( ToRegExpAlgebraic::convert );
auto ToRegExpAlgebraicNFA = registration::AbstractRegister < ToRegExpAlgebraic, regexp::UnboundedRegExp < >, const automaton::NFA < > & > ( ToRegExpAlgebraic::convert );
auto ToRegExpAlgebraicDFA = registration::AbstractRegister < ToRegExpAlgebraic, regexp::UnboundedRegExp < >, const automaton::DFA < > & > ( ToRegExpAlgebraic::convert );

} /* namespace convert */

} /* namespace automaton */
