/*
 * Rename.h
 *
 *  Created on: Dec 9, 2013
 *      Author: Jan Travnicek
 */

#ifndef RENAME_H_
#define RENAME_H_

#include <alib/algorithm>
#include <alib/iostream>
#include <alib/map>
#include <alib/vector>
#include <sstream>

#include <automaton/FSM/DFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/SinglePopDPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>

namespace automaton {

namespace simplify {

class Rename {
public:
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, unsigned > rename ( const automaton::DFA < SymbolType, StateType > & dfa );

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > rename ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda );

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static automaton::SinglePopDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > rename ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda );

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static automaton::InputDrivenDPDA < InputSymbolType, unsigned, unsigned > rename ( const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda );

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static automaton::VisiblyPushdownDPDA < InputSymbolType, unsigned, unsigned > rename ( const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda );

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > rename ( const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda );
};

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, unsigned > Rename::rename ( const automaton::DFA < SymbolType, StateType > & fsm ) {
	unsigned counter = 0;
	ext::map < StateType, unsigned > renamingData;

	for ( const StateType & state : fsm.getStates ( ) )
		renamingData.insert ( std::make_pair ( state, counter++ ) );

	automaton::DFA < SymbolType, unsigned > result ( renamingData.at ( fsm.getInitialState ( ) ) );

	result.setInputAlphabet ( fsm.getInputAlphabet ( ) );

	for ( const StateType & state : fsm.getStates ( ) )
		result.addState ( renamingData.at ( state ) );

	for ( const StateType & state : fsm.getFinalStates ( ) )
		result.addFinalState ( renamingData.at ( state ) );

	for ( const auto & transition : fsm.getTransitions ( ) )
		result.addTransition ( renamingData.at ( transition.first.first ), transition.first.second, renamingData.at ( transition.second ) );

	return result;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > Rename::rename ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda ) {
	unsigned counterState = 0;
	ext::map < StateType, unsigned > renamingDataState;
	unsigned counterSymbol = 0;
	ext::map < InputSymbolType, unsigned > renamingDataSymbol;

	for ( const StateType & state : pda.getStates ( ) )
		renamingDataState.insert ( std::make_pair ( state, counterState++ ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		renamingDataSymbol.insert ( std::make_pair ( symbol, counterSymbol++ ) );

	automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > result ( renamingDataState.at ( pda.getInitialState ( ) ), renamingDataSymbol.at ( pda.getInitialSymbol ( ) ) );

	result.setInputAlphabet ( pda.getInputAlphabet ( ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		result.addPushdownStoreSymbol ( renamingDataSymbol.at ( symbol ) );

	for ( const StateType & state : pda.getStates ( ) )
		result.addState ( renamingDataState.at ( state ) );

	for ( const StateType & state : pda.getFinalStates ( ) )
		result.addFinalState ( renamingDataState.at ( state ) );

	for ( const auto & transition : pda.getTransitions ( ) ) {
		ext::vector < unsigned > pop;

		for ( const InputSymbolType & symbol : std::get < 2 > ( transition.first ) )
			pop.push_back ( renamingDataSymbol.at ( symbol ) );

		ext::vector < unsigned > push;

		for ( const InputSymbolType & symbol : transition.second.second )
			push.push_back ( renamingDataSymbol.at ( symbol ) );

		result.addTransition ( renamingDataState.at ( std::get < 0 > ( transition.first ) ), std::get < 1 > ( transition.first ), pop, renamingDataState.at ( transition.second.first ), push );
	}

	return result;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
automaton::SinglePopDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > Rename::rename ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda ) {
	int counterState = 0;
	ext::map < StateType, unsigned > renamingDataState;
	int counterSymbol = 0;
	ext::map < InputSymbolType, unsigned > renamingDataSymbol;

	for ( const StateType & state : pda.getStates ( ) )
		renamingDataState.insert ( std::make_pair ( state, counterState++ ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		renamingDataSymbol.insert ( std::make_pair ( symbol, counterSymbol++ ) );

	automaton::SinglePopDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > result ( renamingDataState.at ( pda.getInitialState ( ) ), renamingDataSymbol.at ( pda.getInitialSymbol ( ) ) );

	result.setInputAlphabet ( pda.getInputAlphabet ( ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		result.addPushdownStoreSymbol ( renamingDataSymbol.at ( symbol ) );

	for ( const StateType & state : pda.getStates ( ) )
		result.addState ( renamingDataState.at ( state ) );

	for ( const StateType & state : pda.getFinalStates ( ) )
		result.addFinalState ( renamingDataState.at ( state ) );

	for ( const auto & transition : pda.getTransitions ( ) ) {
		ext::vector < unsigned > push;

		for ( const PushdownStoreSymbolType & symbol : transition.second.second )
			push.push_back ( renamingDataSymbol.at ( symbol ) );

		result.addTransition ( renamingDataState.at ( std::get < 0 > ( transition.first ) ), std::get < 1 > ( transition.first ), renamingDataSymbol.at ( std::get < 2 > ( transition.first ) ), renamingDataState.at ( transition.second.first ), push );
	}

	return result;
}

template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
automaton::InputDrivenDPDA < InputSymbolType, unsigned, unsigned > Rename::rename ( const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda ) {
	int counterState = 0;
	ext::map < StateType, unsigned > renamingDataState;
	int counterSymbol = 0;
	ext::map < InputSymbolType, unsigned > renamingDataSymbol;

	for ( const StateType & state : pda.getStates ( ) )
		renamingDataState.insert ( std::make_pair ( state, counterState++ ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		renamingDataSymbol.insert ( std::make_pair ( symbol, counterSymbol++ ) );

	automaton::InputDrivenDPDA < InputSymbolType, unsigned, unsigned > result ( renamingDataState.at ( pda.getInitialState ( ) ), renamingDataSymbol.at ( pda.getInitialSymbol ( ) ) );

	result.setInputAlphabet ( pda.getInputAlphabet ( ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		result.addPushdownStoreSymbol ( renamingDataSymbol.at ( symbol ) );

	for ( const std::pair < const InputSymbolType, std::pair < ext::vector < PushdownStoreSymbolType >, ext::vector < PushdownStoreSymbolType > > > operation : pda.getPushdownStoreOperations ( ) ) {
		ext::vector < unsigned > pop;

		for ( const PushdownStoreSymbolType & symbol : operation.second.first )
			pop.push_back ( renamingDataSymbol.at ( symbol ) );

		ext::vector < unsigned > push;

		for ( const PushdownStoreSymbolType & symbol : operation.second.second )
			push.push_back ( renamingDataSymbol.at ( symbol ) );

		result.setPushdownStoreOperation ( operation.first, pop, push );
	}

	for ( const StateType & state : pda.getStates ( ) )
		result.addState ( renamingDataState.at ( state ) );

	for ( const StateType & state : pda.getFinalStates ( ) )
		result.addFinalState ( renamingDataState.at ( state ) );

	for ( const auto & transition : pda.getTransitions ( ) )
		result.addTransition ( renamingDataState.at ( transition.first.first ), transition.first.second, renamingDataState.at ( transition.second ) );

	return result;
}

template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
automaton::VisiblyPushdownDPDA < InputSymbolType, unsigned, unsigned > Rename::rename ( const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda ) {
	int counterState = 0;
	ext::map < StateType, unsigned > renamingDataState;
	int counterSymbol = 0;
	ext::map < InputSymbolType, unsigned > renamingDataSymbol;

	for ( const StateType & state : pda.getStates ( ) )
		renamingDataState.insert ( std::make_pair ( state, counterState++ ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		renamingDataSymbol.insert ( std::make_pair ( symbol, counterSymbol++ ) );

	automaton::VisiblyPushdownDPDA < InputSymbolType, unsigned, unsigned > result ( renamingDataState.at ( pda.getInitialState ( ) ), renamingDataSymbol.at ( pda.getBottomOfTheStackSymbol ( ) ) );

	result.setCallInputAlphabet ( pda.getCallInputAlphabet ( ) );
	result.setLocalInputAlphabet ( pda.getLocalInputAlphabet ( ) );
	result.setReturnInputAlphabet ( pda.getReturnInputAlphabet ( ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		result.addPushdownStoreSymbol ( renamingDataSymbol.at ( symbol ) );

	for ( const StateType & state : pda.getStates ( ) )
		result.addState ( renamingDataState.at ( state ) );

	for ( const StateType & state : pda.getFinalStates ( ) )
		result.addFinalState ( renamingDataState.at ( state ) );

	for ( const auto & transition : pda.getCallTransitions ( ) )
		result.addCallTransition ( renamingDataState.at ( transition.first.first ), transition.first.second, renamingDataState.at ( transition.second.first ), renamingDataSymbol.at ( transition.second.second ) );

	for ( const auto & transition : pda.getLocalTransitions ( ) )
		result.addLocalTransition ( renamingDataState.at ( transition.first.first ), transition.first.second, renamingDataState.at ( transition.second ) );

	for ( const auto & transition : pda.getReturnTransitions ( ) )
		result.addReturnTransition ( renamingDataState.at ( std::get < 0 > ( transition.first ) ), std::get < 1 > ( transition.first ), renamingDataSymbol.at ( std::get < 2 > ( transition.first ) ), renamingDataState.at ( transition.second ) );

	return result;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > Rename::rename ( const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda ) {
	int counterState = 0;
	ext::map < StateType, unsigned > renamingDataState;
	int counterSymbol = 0;
	ext::map < InputSymbolType, unsigned > renamingDataSymbol;

	for ( const StateType & state : pda.getStates ( ) )
		renamingDataState.insert ( std::make_pair ( state, counterState++ ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		renamingDataSymbol.insert ( std::make_pair ( symbol, counterSymbol++ ) );

	automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, unsigned, unsigned > result ( renamingDataState.at ( pda.getInitialState ( ) ), renamingDataSymbol.at ( pda.getBottomOfTheStackSymbol ( ) ) );

	result.setInputAlphabet ( pda.getInputAlphabet ( ) );

	for ( const InputSymbolType & symbol : pda.getPushdownStoreAlphabet ( ) )
		result.addPushdownStoreSymbol ( renamingDataSymbol.at ( symbol ) );

	for ( const StateType & state : pda.getStates ( ) )
		result.addState ( renamingDataState.at ( state ) );

	for ( const StateType & state : pda.getFinalStates ( ) )
		result.addFinalState ( renamingDataState.at ( state ) );

	for ( const auto & transition : pda.getCallTransitions ( ) )
		result.addCallTransition ( renamingDataState.at ( transition.first.first ), transition.first.second, renamingDataState.at ( transition.second.first ), renamingDataSymbol.at ( transition.second.second ) );

	for ( const auto & transition : pda.getLocalTransitions ( ) )
		result.addLocalTransition ( renamingDataState.at ( transition.first.first ), transition.first.second, renamingDataState.at ( transition.second ) );

	for ( const auto & transition : pda.getReturnTransitions ( ) )
		result.addReturnTransition ( renamingDataState.at ( std::get < 0 > ( transition.first ) ), std::get < 1 > ( transition.first ), renamingDataSymbol.at ( std::get < 2 > ( transition.first ) ), renamingDataState.at ( transition.second ) );

	return result;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* RENAME_H_ */
