/*
 * SingleInitialStateEpsilonTransition.h
 *
 *  Created on: 6. 6. 2018
 *	  Author: Jan Travnicek
 */

#ifndef _SINGLE_INITIAL_STATE_H_
#define _SINGLE_INITIAL_STATE_H_

#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/MultiInitialStateEpsilonNFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/CompactNFA.h>

#include <label/InitialStateLabel.h>
#include <common/createUnique.hpp>

namespace automaton {

namespace simplify {

/**
 * Makes finite automaton's transition function convert.
 * Source: Melichar: Algorithm 2.22
 */
class SingleInitialStateEpsilonTransition {
public:
	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	template < class SymbolType, class StateType >
	static automaton::EpsilonNFA < SymbolType, DefaultEpsilonType, StateType > convert(const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > convert(const automaton::MultiInitialStateEpsilonNFA < SymbolType, EpsilonType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > convert(const automaton::DFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > convert(const automaton::NFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > convert(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static automaton::ExtendedNFA < SymbolType, StateType > convert(const automaton::ExtendedNFA < SymbolType, StateType > & automaton);
	template < class SymbolType, class StateType >
	static automaton::CompactNFA < SymbolType, StateType > convert(const automaton::CompactNFA < SymbolType, StateType > & automaton);
};

template < class SymbolType, class StateType >
automaton::EpsilonNFA < SymbolType, DefaultEpsilonType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton) {
	// step 1, 3
	StateType q0 = common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), automaton.getStates ( ) );

	automaton::EpsilonNFA < SymbolType, DefaultEpsilonType, StateType > res ( q0 );
	res.setInputAlphabet ( automaton.getInputAlphabet ( ) );

	for( const StateType & q : automaton.getStates ( ) )
		res.addState( q );

	// step 2
	for ( const auto & q : automaton.getInitialStates ( ) )
		res.addTransition ( q0, q );

	for ( const auto & t : automaton.getTransitions ( ) )
		for ( const auto & q : t.second )
			res.addTransition ( t.first.first, t.first.second, q );

	// step 4, 5
	res.setFinalStates ( automaton.getFinalStates ( ) );

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::MultiInitialStateEpsilonNFA < SymbolType, EpsilonType, StateType > & automaton) {
	// step 1, 3
	StateType q0 = common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), automaton.getStates ( ) );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res ( q0 );
	res.setInputAlphabet ( automaton.getInputAlphabet ( ) );

	for( const StateType & q : automaton.getStates ( ) )
		res.addState( q );

	// step 2
	for ( const auto & q : automaton.getInitialStates ( ) )
		res.addTransition ( q0, q );

	for ( const auto & t : automaton.getTransitions ( ) )
		for ( const StateType & q : t.second )
			res.addTransition ( t.first.first, t.first.second, q );

	// step 4, 5
	res.setFinalStates ( automaton.getFinalStates ( ) );

	return res;
}

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::DFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::NFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::CompactNFA < SymbolType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::CompactNFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::ExtendedNFA < SymbolType, StateType > SingleInitialStateEpsilonTransition::convert(const automaton::ExtendedNFA < SymbolType, StateType > & automaton) {
	return automaton;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* _SINGLE_INITIAL_STATE_H_ */
