/*
 * AutomataConcatenation.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_CONCATENATION_H_
#define AUTOMATA_CONCATENATION_H_

#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Concatenation of two finite automata.
 * For finite automata A1, A2, we create a finite automaton A such that L(A) = L(A1).L(A2).
 * This method utilizes epsilon transitions in finite automata (Melichar: Jazyky a překlady, 2.82).
 */
class AutomataConcatenation {
public:
	/**
	 * Concatenates two finite automata without using epsilon transitions.
	 * @tparam SymbolType Type for input symbols.
	 * @tparam StateType Type for states.
	 * @param first First automaton (A1)
	 * @param second Second automaton (A2)
	 * @return nondeterministic FA representing the concatenation of two automata
	 */
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second);
};

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > AutomataConcatenation::concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	StateType q01q02(first.getInitialState(), second.getInitialState());
	automaton::NFA < SymbolType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		for(const auto& q : t.second) {
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

			if(first.getFinalStates().count(q) > 0)
				res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(secondDefault, second.getInitialState()));
		}
	}
	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, StateType(firstDefault, q));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > AutomataConcatenation::concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	StateType q01q02(first.getInitialState(), second.getInitialState());
	automaton::NFA < SymbolType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, t.second));

		if(first.getFinalStates().count(t.second) > 0)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(secondDefault, second.getInitialState()));
	}
	for(const auto& t : second.getTransitions())
		res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, t.second));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		res.addTransition(q01q02, t.first.second, StateType(firstDefault, t.second));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		res.addTransition(q01q02, t.first.second, StateType(secondDefault, t.second));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_CONCATENATION_H_ */
