/*
 * AutomataConcatenationEpsilonTransition.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_CONCATENATION_EPSILON_TRANSITION_H_
#define AUTOMATA_CONCATENATION_EPSILON_TRANSITION_H_

#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Concatenation of two finite automata.
 * For finite automata A1, A2, we create a finite automaton A such that L(A) = L(A1).L(A2).
 * This method utilizes epsilon transitions in finite automata (Melichar: Jazyky a překlady, 2.80).
 */
class AutomataConcatenationEpsilonTransition {
public:
	/**
	 * Concatenates two automata using epsilon transitions.
	 * @tparam SymbolType Type for input symbols.
	 * @tparam EpsilonType Type for epsilon symbol. Defaults to DefaultEpsilonType.
	 * @tparam StateType Type for states.
	 * @param first First automaton (A1)
	 * @param second Second automaton (A2)
	 * @return nondeterministic FA with epsilon transitions representing the concatenation of two automata
	 */
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second);

	/**
	 * @overload
	 */
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second);

	/**
	 * @overload
	 */
	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > concatenation(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second);
};

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, t.second));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataConcatenationEpsilonTransition::concatenation(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& q : first.getFinalStates())
		res.addTransition(StateType(firstDefault, q), StateType(secondDefault, second.getInitialState()));

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_CONCATENATION_EPSILON_H_ */
