/*
 * ToAutomatonDerivation.h
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#ifndef TO_AUTOMATON_DERIVATION_H_
#define TO_AUTOMATON_DERIVATION_H_

#include <regexp/formal/FormalRegExp.h>
#include <regexp/unbounded/UnboundedRegExp.h>

#include <automaton/FSM/NFA.h>

#include <regexp/RegExp.h>

#include <alib/set>
#include <alib/deque>
#include <alib/vector>

#include <string/LinearString.h>

#include <regexp/transform/RegExpDerivation.h>
#include <regexp/simplify/RegExpOptimize.h>
#include <regexp/properties/RegExpEpsilon.h>

namespace regexp {

namespace convert {

/**
 * Converts regular expression to finite automaton using BrzozowskiDerivation algorithm (derivations of regular expressions).
 * Source: Melichar 2.110
 */
class ToAutomatonDerivation {
public:
	/**
	 * Performs conversion.
	 * @return FSM equivalent to original regular expression.
	 */
	template < class T, class SymbolType = typename T::symbol_type >
	static automaton::DFA < SymbolType, unsigned > convert ( const T & regexp );
};

template < class T, class SymbolType >
automaton::DFA < SymbolType, unsigned > ToAutomatonDerivation::convert(const T& regexp) {
	// 1.
	T V = regexp::simplify::RegExpOptimize::optimize(regexp);

	ext::set<T> Q = { V };
	ext::deque<ext::set<T>> Qi;

	Qi.push_back(ext::set<T>());
	Qi.at(0).insert(V);

	int i = 1;

	// 2.
	while(! Qi.at(i - 1).empty()) {
		Qi.push_back(ext::set<T>()); // initialize set Q_i

		for(const auto& dregexp : Qi.at(i - 1)) {

			for(const auto& a : regexp.getAlphabet()) {
				string::LinearString < > string(ext::vector<DefaultSymbolType>{a});
				T derived = regexp::RegExpDerivation::derivation(dregexp, string);
				derived = regexp::simplify::RegExpOptimize::optimize(derived);

				// this will also add \emptyset as a regexp (and as FA state)
				if(Q.count(derived) == 0) // if this state has already been found, do not add
					Qi.at(i).insert(derived);
			}
		}

		Q.insert(Qi.at(i).begin(), Qi.at(i).end());
		i += 1;
	}

	ext::map<T, unsigned> stateMap;
	unsigned stateId = 0;

	for(const auto& r : Q) {
		stateMap.insert ( std::make_pair ( r, stateId ++ ) );
	}

	// ------------------------------------------------------------------------
	// 3.

	automaton::DFA < SymbolType, unsigned > automaton ( stateMap.find ( V )->second );

	for(const auto& r : stateMap) {
		automaton.addState(r.second);
	}

	automaton.setInputAlphabet(regexp.getAlphabet());

	for(const auto& r : Q) {

		for(const auto& a: regexp.getAlphabet()) {
			string::LinearString < > string(ext::vector<DefaultSymbolType>{a});
			T derived = regexp::RegExpDerivation::derivation(r, string);
			derived = regexp::simplify::RegExpOptimize::optimize(derived);

			automaton.addTransition(stateMap.find(r)->second, a, stateMap.find(derived)->second);
		}
	}

	for(const auto& r : Q) {
		if(regexp::properties::RegExpEpsilon::languageContainsEpsilon(r))
			automaton.addFinalState(stateMap.find(r)->second);
	}

	return automaton;
}

} /* namespace convert */

} /* namespace regexp */

#endif /* TO_AUTOMATON_DERIVATION_H_ */
