/*
 * Deque.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_DEQUE_H_
#define OBJECTS_DEQUE_H_

#include <alib/deque>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>

namespace container {

/**
 * \brief
 * Represents an adaptor of a deque container from the stl extensions provided by the algorithms library.
 *
 * The deque is a linear random access container.
 *
 * \tparam ElementType the type of values stored in the container.
 */
template < class ElementType = object::Object >
class ObjectsDeque final : public ext::deque < ElementType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the container based on the extended version of the stl deque container.
	 *
	 * \param raw the extended deque container
	 */
	explicit ObjectsDeque ( ext::deque < ElementType > raw );

	/**
	 * \brief
	 * Creates a new instance of the contaier based on defaultly construted underlying container container.
	 */
	explicit ObjectsDeque ( );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsDeque & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( ext::deque < ElementType > raw ) : ext::deque < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( ) : ext::deque < ElementType > ( ) {
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::clone ( ) const & {
	return new ObjectsDeque ( * this );
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::clone ( ) && {
	return new ObjectsDeque ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsDeque < ElementType >::compare ( const ObjectsDeque & other ) const {
	static ext::compare < ext::deque < ElementType > > comp;

	return comp ( static_cast < const ext::deque < ElementType > & > ( * this ), static_cast < const ext::deque < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsDeque < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsDeque " << static_cast < const ext::deque < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsDeque < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
object::ObjectBase* ObjectsDeque < ElementType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class ElementType >
struct normalize < container::ObjectsDeque < ElementType > > {
	static ext::deque < object::Object > raw ( ext::deque < ElementType > && source ) {
		ext::deque < object::Object > res;

		for ( ElementType & element : source )
			res.push_back ( object::ObjectFactory::make ( std::move ( element ) ) );

		return res;
	}

	static container::ObjectsDeque < > eval ( container::ObjectsDeque < ElementType > && value ) {
		return container::ObjectsDeque < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_DEQUE_H_ */
