/*
 * Map.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_MAP_H_
#define OBJECTS_MAP_H_

#include <alib/map>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>
#include <container/ObjectsPair.h>

namespace container {

/**
 * \brief
 * Represents an adaptor of a map container from the stl extensions provided by the algorithms library.
 *
 * The map is an associative container.
 *
 * \tparam KeyType the type of keys in the map.
 * \tparam ValueType the type of values in the map.
 */
template < class KeyType = object::Object, class ValueType = object::Object >
class ObjectsMap final : public ext::map < KeyType, ValueType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the container based on the extended version of the stl map container.
	 *
	 * \param raw the extended map container
	 */
	explicit ObjectsMap ( ext::map < KeyType, ValueType > raw );

	/**
	 * \brief
	 * Creates a new instance of the contaier based on defaultly construted underlying container.
	 */
	explicit ObjectsMap ( );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsMap & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap ( ext::map < KeyType, ValueType > raw ) : ext::map < KeyType, ValueType > ( std::move ( raw ) ) {

}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap() : ext::map<KeyType, ValueType>() {

}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::clone ( ) const & {
	return new ObjectsMap(*this);
}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::clone() && {
	return new ObjectsMap(std::move(*this));
}

template < class KeyType, class ValueType >
int ObjectsMap<KeyType, ValueType>::compare(const ObjectsMap& other) const {
	static ext::compare<ext::map<KeyType, ValueType>> comp;
	return comp ( static_cast < const ext::map < KeyType, ValueType > & > ( * this ), static_cast < const ext::map < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsMap<KeyType, ValueType>::operator>>(std::ostream& os) const {
	os << "(ObjectsMap " << static_cast < const ext::map < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class KeyType, class ValueType >
object::ObjectBase* ObjectsMap < KeyType, ValueType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class KeyType, class ValueType >
struct normalize < container::ObjectsMap < KeyType, ValueType > > {
	static ext::map < object::Object, object::Object > raw ( ext::map < KeyType, ValueType > && source ) {
		ext::map < object::Object, object::Object > res;

		for ( std::pair < KeyType, ValueType > && element : ext::make_moveable_map ( source ) )
			res.insert ( normalize < container::ObjectsPair < KeyType, ValueType > >::raw2 ( std::move ( element ) ) );

		return res;
	}

	static container::ObjectsMap < > eval ( container::ObjectsMap < KeyType, ValueType > && value ) {
		return container::ObjectsMap < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_MAP_H_ */
