/*
 * Set.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_SET_H_
#define OBJECTS_SET_H_

#include <alib/set>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>

namespace container {

/**
 * \brief
 * Represents an adaptor of a set container from the stl extensions provided by the algorithms library.
 *
 * The set is an associative container.
 *
 * \tparam ElementType the type of values stored in the container.
 */
template < class ElementType = object::Object >
class ObjectsSet final : public ext::set < ElementType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the container based on the extended version of the stl set container.
	 *
	 * \param raw the extended set contaier
	 */
	explicit ObjectsSet ( ext::set < ElementType > raw );

	/**
	 * \brief
	 * Creates a new instance of the contaier based on defaultly construted underlying container.
	 */
	explicit ObjectsSet ( );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsSet & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet(ext::set<ElementType> raw) : ext::set<ElementType>(std::move(raw)) {

}

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet() : ext::set<ElementType>() {

}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::clone ( ) const & {
	return new ObjectsSet(*this);
}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::clone() && {
	return new ObjectsSet(std::move(*this));
}

template < class ElementType >
int ObjectsSet < ElementType >::compare(const ObjectsSet& other) const {
	static ext::compare<ext::set<ElementType>> comp;
	return comp ( static_cast < const ext::set < ElementType > & > ( * this ), static_cast < const ext::set < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsSet < ElementType >::operator>>(std::ostream& os) const {
	os << "(ObjectsSet " << static_cast < const ext::set < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsSet < ElementType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class ElementType >
object::ObjectBase* ObjectsSet < ElementType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class ElementType >
struct normalize < container::ObjectsSet < ElementType > > {
	static ext::set < object::Object > raw ( ext::set < ElementType > && source ) {
		ext::set < object::Object > res;

		for ( ElementType && element : ext::make_moveable_set ( source ) )
			res.insert ( object::ObjectFactory::make ( std::move ( element ) ) );

		return res;
	}

	static container::ObjectsSet < > eval ( container::ObjectsSet < ElementType > && value ) {
		return container::ObjectsSet < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_SET_H_ */
