/*
 * ObjectsTrie.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TRIE_H_
#define OBJECTS_TRIE_H_

#include <alib/trie>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>

namespace container {

/**
 * \brief
 * Represents an adaptor of a trie container from the stl extensions provided by the algorithms library.
 *
 * The trie is an structured associative container.
 *
 * \tparam KeyType the type of keys to access subnodes of a trie node.
 * \tparam ValueType the type of values in the trie nodes.
 */
template < class KeyType = object::Object, class ValueType = object::Object >
class ObjectsTrie final : public ext::trie < KeyType, ValueType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the trie container wrapper based on the tree datatype from stl extensions of the algorithms library.
	 *
	 * \param raw the extended tree contaier
	 */
	explicit ObjectsTrie ( ext::trie < KeyType, ValueType > raw );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsTrie & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::ObjectsTrie ( ext::trie < KeyType, ValueType > raw ) : ext::trie < KeyType, ValueType > ( std::move ( raw ) ) {
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::clone ( ) const & {
	return new ObjectsTrie ( * this );
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::clone ( ) && {
	return new ObjectsTrie ( std::move ( * this ) );
}

template < class KeyType, class ValueType >
int ObjectsTrie < KeyType, ValueType >::compare ( const ObjectsTrie & other ) const {
	static ext::compare < ext::trie < KeyType, ValueType > > comp;

	return comp ( static_cast < const ext::trie < KeyType, ValueType > & > ( * this ), static_cast < const ext::trie < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTrie " << static_cast < const ext::trie < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class KeyType, class ValueType >
object::ObjectBase* ObjectsTrie < KeyType, ValueType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class KeyType, class ValueType >
struct normalize < container::ObjectsTrie < KeyType, ValueType > > {
	static ext::trie < object::Object, object::Object > raw ( ext::trie < KeyType, ValueType > && source ) {
		ext::map < object::Object, ext::trie < object::Object, object::Object > > children;

		for ( std::pair < KeyType, ext::trie < KeyType, ValueType > > && child : ext::make_moveable_map ( source.getChildren ( ) ) ) {
			children.insert ( std::make_pair ( object::ObjectFactory::make ( std::move ( child.first ) ), raw ( std::move ( child.second ) ) ) );
		}

		return ext::trie < object::Object, object::Object > ( object::ObjectFactory::make ( std::move ( source.getData ( ) ) ), std::move ( children ) );
	}

	static container::ObjectsTrie < > eval ( container::ObjectsTrie < KeyType, ValueType > && value ) {
		return container::ObjectsTrie < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_TRIE_H_ */
