/*
 * ObjectFactory.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Mar 16, 2018
 *      Author: Jan Travnicek
 */

#ifndef _OBJECT_FACTORY_H__
#define _OBJECT_FACTORY_H__

#include "Object.h"
#include <alib/variant>
#include <object/AnyObject.h>
#include <object/Void.h>

namespace object {

/**
 * \brief
 * Class representing uniqfied object construction.
 *
 * \details
 * The class provides a static make method that is overloaded for some primitive types, variant, etc...
 */
class ObjectFactory {
	/**
	 * \brief
	 * Variant access method to create object from variant containing void.
	 */
	template < class Variant, class Type, typename std::enable_if < std::is_same < Type, void >::value, Type * >::type * = nullptr >
	static Object bypassVariant ( Variant && ) {
		return Object ( Void { } );
	}

	/**
	 * \brief
	 * Variant access method to create object from variant containing something else than void.
	 */
	template < class Variant, class Type, typename std::enable_if < ! std::is_same < Type, void >::value, Type * >::type * = nullptr >
	static Object bypassVariant ( Variant && data ) {
		return make ( std::move ( data.template get < Type > ( ) ) );
	}

	/**
	 * \brief
	 * Helper method to detect variant content and call make method on the object from inside of the variant.
	 */
	template < class Variant, class Type, class ... Types >
	static Object processVariant ( Variant && data ) {
		if ( data.template is < Type > ( ) ) {
			return bypassVariant < Variant, Type > ( std::move ( data ) );
		} else {
			return processVariant < Variant, Types ... > ( std::move ( data ) );
		}
	}

	/**
	 * \brief
	 * Root case of the variant traversing recursion.
	 */
	template < class Variant >
	static Object processVariant ( Variant && ) {
		throw std::logic_error ( "Variant state error" );
	}

public:
	/**
	 * \brief
	 * Specialisation of the make method for booleans.
	 */
	static Object make ( bool boolean );

	/**
	 * \brief
	 * Specialisation of the make method for charactes.
	 */
	static Object make ( char character );

	/**
	 * \brief
	 * Specialisation of the make method for doubles.
	 */
	static Object make ( double number );

	/**
	 * \brief
	 * Specialisation of the make method for integers.
	 */
	static Object make ( int number );

	/**
	 * \brief
	 * Specialisation of the make method for strings.
	 */
	static Object make ( std::string string );

	/**
	 * \brief
	 * Specialisation of the make method for c-strings.
	 */
	static Object make ( const char * string );

	/**
	 * \brief
	 * Specialisation of the make method for unsigneds.
	 */
	static Object make ( unsigned number );

	/**
	 * \brief
	 * Specialisation of the make method for rvalue referenced objects.
	 */
	static Object && make ( Object && other ) {
		return std::move ( other );
	}

	/**
	 * \brief
	 * Specialisation of the make method for lvalue referenced objects.
	 */
	static Object & make ( Object & other ) {
		return other;
	}

	/**
	 * \brief
	 * Specialisation of the make method for constant lvalue referenced objects.
	 */
	static const Object & make ( const Object & other ) {
		return other;
	}

	/**
	 * \brief
	 * Specialisation of the make method for objects that are not from the object hierarchy of Algorithms library
	 */
	template < class Type, typename std::enable_if < ! std::is_convertible < Type *, object::ObjectBase * >::value, Type >::type * = nullptr >
	static Object make ( Type && data ) {
		return Object ( object::AnyObject < Type > ( std::forward < Type > ( data ) ) );
	}

	/**
	 * \brief
	 * Specialisation of the make method for objects that are from the object hierarchy of Algorithms library
	 */
	static Object make ( object::ObjectBase && data ) {
		return Object ( std::move ( data ) );
	}

	/**
	 * \brief
	 * Specialisation of the make method for objects that are from the object hierarchy of Algorithms library
	 */
	static Object make ( const object::ObjectBase & data ) {
		return Object ( data );
	}

	/**
	 * \brief
	 * Specialisation of the make method for variants.
	 *
	 * \details
	 * The resulting object is not constructed from the variant itself but from the value stored inside. If the value stored inside the variant is variant again, the process is repeated.
	 */
	template < class ... Types >
	static Object make ( ext::variant < Types ... > && data ) {
		return processVariant < ext::variant < Types ... >, Types ... > ( std::move ( data ) );
	}
};

} /* namespace object */

#endif /* _OBJECT_FACTORY_H__ */
