/*
 * BottomOfTheStackSymbol.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Jun 19, 2014
 *      Author: Jan Travnicek
 */

#ifndef BOTTOM_OF_THE_STACK_SYMBOL_H_
#define BOTTOM_OF_THE_STACK_SYMBOL_H_

#include "SymbolBase.h"
#include "Symbol.h"
#include <object/Object.h>
#include <common/ranked_symbol.hpp>

namespace alphabet {

/**
 * \brief
 * Represents bottom of the stack symbol used in the visibly pushdown automata.
 */
class BottomOfTheStackSymbol : public SymbolBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the symbol.
	 */
	explicit BottomOfTheStackSymbol ( );

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) const &
	 */
	virtual SymbolBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::SymbolBase::clone ( ) &&
	 */
	virtual SymbolBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same symbols
	 */
	int compare ( const BottomOfTheStackSymbol & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if <   std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * \brief Factory for the symbol construction of the symbol based on given type.
	 */
	template < typename Base >
	static inline typename std::enable_if < ! std::is_integral < Base >::value, Base >::type instance ( );

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < typename Base >
inline typename std::enable_if < std::is_integral < Base >::value, Base >::type BottomOfTheStackSymbol::instance ( ) {
	return '_';
}

template < >
inline object::Object BottomOfTheStackSymbol::instance < object::Object > ( ) {
	return object::Object ( BottomOfTheStackSymbol ( ) );
}

template < >
inline std::string BottomOfTheStackSymbol::instance < std::string > ( ) {
	return std::string ( 1, BottomOfTheStackSymbol::instance < char > ( ) );
}

template < >
inline BottomOfTheStackSymbol BottomOfTheStackSymbol::instance < BottomOfTheStackSymbol > ( ) {
	return BottomOfTheStackSymbol ( );
}

template < >
inline Symbol BottomOfTheStackSymbol::instance < Symbol > ( ) {
	return Symbol ( BottomOfTheStackSymbol ( ) );
}

template < >
inline common::ranked_symbol < > BottomOfTheStackSymbol::instance < common::ranked_symbol < > > ( ) {
	return common::ranked_symbol < > ( BottomOfTheStackSymbol::instance < DefaultSymbolType > ( ), DefaultRankType ( 0 ) );
}

} /* namespace alphabet */

#endif /* BOTTOM_OF_THE_STACK_SYMBOL_H_ */
