/*
 * forward_list.hpp
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef __FORWARD_LIST_HPP_
#define __FORWARD_LIST_HPP_

#include <forward_list>
#include <ostream>
#include <sstream>
#include <string>

#include "compare.hpp"
#include "allocFix.hpp"

namespace ext {

/**
 * Class extending the forward_list class from the standard library. Original reason is to allow printing of the container with overloaded operator <<.
 *
 * The class mimics the behavior of the forward_list from the standatd library.
 *
 * \tparam T the type of values inside the forward_list
 * \tparam Alloc the allocator of values of type T
 */
template < class T, class Alloc = std::allocator < T > >
class forward_list : public std::forward_list < T, Alloc >, AllocFix < Alloc > {
public:
	/**
	 * Inherit constructors of the standard forward_list
	 */
	using std::forward_list < T, Alloc >::forward_list;

	/**
	 * Inherit operator = of the standard forward_list
	 */
	using std::forward_list < T, Alloc >::operator =;
#ifndef __clang__

	/**
	 * Default constructor needed by g++ since it is not inherited
	 */
	forward_list ( ) = default;

	/**
	 * Copy constructor needed by g++ since it is not inherited
	 */
	forward_list ( const forward_list & other ) = default;

	/**
	 * Move constructor needed by g++ since it is not inherited
	 */
	forward_list ( forward_list && other ) = default;

	/**
	 * Copy operator = needed by g++ since it is not inherited
	 */
	forward_list & operator = ( forward_list && other ) = default;

	/**
	 * Move operator = needed by g++ since it is not inherited
	 */
	forward_list & operator = ( const forward_list & other ) = default;
#endif
};

/**
 * Operator to print the forward_list to the output stream.
 *
 * \param out the output stream
 * \param forward_list the forward_list to print
 *
 * \tparam T the type of values inside the forward_list
 * \tparam Ts ... remaining unimportant template parameters of the forward_list
 *
 * \return the output stream from the \p out
 */
template< class T, class ... Ts >
std::ostream& operator<<(std::ostream& out, const ext::forward_list<T, Ts ... >& forward_list) {
	out << "[";

	bool first = true;
	for(const T& item : forward_list) {
		if(!first) out << ", ";
		first = false;
		out << item;
	}

	out << "]";
	return out;
}

/**
 * Specialisation of the compare structure implementing the three-way comparison
 *
 * \tparam T the type of values inside the forward_list
 * \tparam Ts ... remaining unimportant template parameters of the forward_list
 */
template<class T, class ... Ts >
struct compare<ext::forward_list<T, Ts ... >> {

	/**
	 * Implementation of the three-way comparison
	 *
	 * \param first the left operand of the comparison
	 * \param second the right operand of the comparison
	 *
	 * \return negative value of left < right, positive value if left > right, zero if left == right
	 */
	int operator()(const ext::forward_list<T, Ts ... >& first, const ext::forward_list<T, Ts ... >& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		static compare<typename std::decay < T >::type > comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); ++iterF, ++iterS) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

/**
 * Overload of to_string function.
 *
 * \param value the forward_list to be converted to string
 *
 * \tparam T the type of values inside the forward_list
 * \tparam Ts ... remaining unimportant template parameters of the forward_list
 */
template < class T, class ... Ts >
std::string to_string ( const ext::forward_list < T, Ts ... > & value ) {
	std::stringstream ss;
	ss << value;
	return ss.str();
}

} /* namespace ext */

#endif /* __FORWARD_LIST_HPP_ */

