/*
 * iterator.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef __ITERATOR_HPP_
#define __ITERATOR_HPP_

#include <iterator>

namespace ext {

template<class T>
class reverser {
	T & m_Container;

public:
	reverser ( T & container ) : m_Container ( container ) {
	}

	typename T::reverse_iterator begin ( ) const {
		return m_Container.rbegin();
	}

	typename T::reverse_iterator end ( ) const {
		return m_Container.rend();
	}
};

template < class T >
class reverser < const T > {
	const T & m_Container;

public:
	reverser ( const T & container ) : m_Container ( container ) {
	}

	typename T::const_reverse_iterator begin ( ) const {
		return m_Container.rbegin();
	}

	typename T::const_reverse_iterator end ( ) const {
		return m_Container.rend();
	}
};

template<class T>
reverser<T> make_reverse(T& container) {
	return reverser<T>(container);
}

template < class Iterator >
class dereferencing_iterator {
	Iterator m_base;

public:
	using pointer = typename std::iterator_traits < Iterator >::value_type;
	using value_type = typename std::remove_pointer < pointer >::type;
	using difference_type = std::ptrdiff_t;
	using reference = value_type &;
	using iterator_category = typename std::iterator_traits < Iterator >::iterator_category;

	explicit dereferencing_iterator ( Iterator base ) : m_base ( base ) {
	}

	template < class Iter >
	dereferencing_iterator ( const dereferencing_iterator < Iter > & iter ) : m_base ( iter.base ( ) ) {
	}

	// Dereferencable.
	reference operator * ( ) const {
		return * * m_base;
	}

	pointer operator -> ( ) const {
		return * m_base;
	}

	// Pre- and post-incrementable.
	dereferencing_iterator < Iterator > & operator ++ ( ) {
		++ m_base;
		return * this;
	}

	dereferencing_iterator < Iterator > operator ++ ( int ) {
		auto tmp = * this;
		++ m_base;
		return tmp;
	}

	// Pre- and post-decrementable.
	dereferencing_iterator < Iterator > & operator -- ( ) {
		++ m_base;
		return * this;
	}

	dereferencing_iterator < Iterator > operator -- ( int ) {
		auto tmp = * this;
		++ m_base;
		return tmp;
	}

	// Add and substract.
	dereferencing_iterator < Iterator > & operator += ( int distance ) {
		m_base += distance;
		return *this;
	}

	dereferencing_iterator < Iterator > operator + ( int distance ) const {
		auto res = * this;
		res += distance;
		return res;
	}

	dereferencing_iterator < Iterator > & operator -= ( int distance ) {
		m_base -= distance;
		return *this;
	}

	dereferencing_iterator < Iterator > operator - ( int distance ) const {
		auto res = * this;
		res -= distance;
		return res;
	}

	//Array subscript
	reference operator [ ] ( int index ) const {
		return * ( m_base [ index ] );
	}

	difference_type operator - ( const dereferencing_iterator < Iterator > & other ) {
		return m_base - other.m_base;
	}

	// Equality / inequality.
	bool operator == ( const dereferencing_iterator < Iterator > & other ) const {
		return this->m_base == other.m_base;
	}

	bool operator != ( const dereferencing_iterator < Iterator > & other ) const {
		return ! ( * this == other );
	}

	bool operator < ( const dereferencing_iterator < Iterator > & other ) const {
		return this->m_base < other.m_base;
	}

	bool operator > ( const dereferencing_iterator < Iterator > & other ) const {
		return other < * this;
	}

	bool operator <= ( const dereferencing_iterator < Iterator > & other ) const {
		return ! ( * this > other );
	}

	bool operator >= ( const dereferencing_iterator < Iterator > & other ) const {
		return ! ( * this < other );
	}

	Iterator base ( ) const {
		return m_base;
	}

};

template < class Iterator >
dereferencing_iterator < Iterator > dereferencer ( Iterator iter ) {
	return dereferencing_iterator < Iterator > ( iter );
}

template < typename T >
struct is_iterator : std::false_type {};

template < typename T >
struct is_iterator<typename std::iterator_traits<T>> : std::true_type {};

template < typename BidirectionalIterator, typename Distance >
inline constexpr void retractInternal ( BidirectionalIterator & i, Distance n, std::bidirectional_iterator_tag ) {
	while ( n -- )
		-- i;
	while ( n ++ )
		++ i;
}

template < typename RandomAccessIterator, typename Distance >
inline constexpr void retractInternal ( RandomAccessIterator & i, Distance n, std::random_access_iterator_tag ) {
	i += n;
}

/**
 *  \brief A generalization of pointer arithmetic.
 *  \param i A bidirectional iterator.
 *  \param n The \a delta by which to change \p i.
 *
 *  This decrements \p i by \p n. For bidirectional and random access
 *  iterators, \p n may be negative, in which case \p i is incremented.
 *
 *  For random access iterators, this uses their \c + and \c - operations
 *  and are constant time. For other %iterator classes they are linear time.
 */
template < typename BidirectionalIterator, typename Distance >
inline constexpr void retract ( BidirectionalIterator & i, Distance n ) {
	typename std::iterator_traits < BidirectionalIterator >::difference_type d = n;
	retractInternal ( i, d, std::__iterator_category ( i ) );
}

} /* namespace ext */

#endif /* __ITERATOR_HPP_ */
