/*
 * map.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef __MAP_HPP_
#define __MAP_HPP_

#include <map>
#include <ostream>
#include <sstream>
#include <utility>
#include <string>

#include "compare.hpp"
#include "pair.hpp"
#include "allocFix.hpp"

namespace ext {

template < typename T, typename R, typename Cmp = std::less < >, typename Alloc = std::allocator < std::pair < const T, R > > >
class map : public std::map < T, R, Cmp, Alloc >, AllocFix < Alloc > {
public:
	using std::map< T, R, Cmp, Alloc >::map;
	using std::map< T, R, Cmp, Alloc >::operator =;
#ifndef __clang__
	map ( ) = default;

	map ( const map & other ) = default;

	map ( map && other ) = default;

	map & operator = ( map && other ) = default;

	map & operator = ( const map & other ) = default;
#endif
	using iterator = typename std::map<T, R, Cmp, Alloc>::iterator;
	using key_type = typename std::map<T, R, Cmp, Alloc>::key_type;

	// Follow interface from C++17
	// TODO: Remove this member function after move to C++17
	template < typename M >
	pair < iterator, bool > insert_or_assign ( const key_type & k, M && obj ) {
		iterator pos = this->find ( k );
		if ( pos == this->end ( ) ) {
			pos = this->insert ( ext::make_pair ( k, std::forward < M > ( obj ) ) ).first;
			return ext::make_pair ( pos, true );
		} else {
			pos->second = std::forward < M > ( obj );
			return ext::make_pair ( pos, false );
		}
	}

};

template< class T, class R, class ... Ts >
std::ostream& operator<<(std::ostream& out, const ext::map<T, R, Ts ... >& map) {
	out << "{";

	bool first = true;
	for(const std::pair<const T, R>& item : map) {
		if(!first) out << ", ";
		first = false;
		out << "(" << item.first << ", " << item.second << ")";
	}

	out << "}";
	return out;
}

template < class T, class R, class ... Ts >
struct compare < ext::map < T, R, Ts ... > > {
	int operator ( ) ( const ext::map < T, R, Ts ... > & first, const ext::map < T, R, Ts ... > & second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		static compare < typename std::decay < T >::type > compT;
		static compare < typename std::decay < R >::type > compR;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); ++iterF, ++iterS) {
			int res = compT(iterF->first, iterS->first);
			if(res == 0) res = compR(iterF->second, iterS->second);
			if(res != 0) return res;
		}
		return 0;
	}
};

template < class T, class R, class ... Ts >
std::string to_string ( const ext::map < T, R, Ts ... > & value ) {
	std::stringstream ss;
	ss << value;
	return ss.str();
}

template<typename map_type>
class key_iterator : public map_type::const_iterator {
public:
	typedef typename map_type::const_iterator map_iterator;
	typedef typename map_iterator::value_type::first_type key_type;

	key_iterator(const map_iterator& other) : map_iterator(other) {} ;

	const key_type& operator *() const {
		return map_iterator::operator*().first;
	}

	const key_type* operator ->() const {
		return &(map_iterator::operator*().first);
	}
};

// helpers to create iterators easier:
template<typename map_type>
key_iterator<map_type> key_begin(const map_type& m) {
	return key_iterator<map_type>(m.begin());
}

template<typename map_type>
key_iterator<map_type> key_end(const map_type& m) {
	return key_iterator<map_type>(m.end());
}

template <class Iterator, class KeyType, class ValueType>
class map_move_iterator {
	Iterator current;

public:
	typedef Iterator iterator_type;
	typedef typename std::iterator_traits<Iterator>::iterator_category iterator_category;
	typedef typename std::pair < KeyType, ValueType >        value_type;
	typedef Iterator pointer;
	typedef value_type&& reference;

	explicit map_move_iterator() {
	}

	explicit map_move_iterator (Iterator it) : current(it) {
	}

	map_move_iterator (const map_move_iterator<Iterator, KeyType, ValueType>& it) : current(it.current) {
	}

	map_move_iterator& operator= (const map_move_iterator<Iterator, KeyType, ValueType>& it) {
		current = it.current;
	}

	iterator_type base() const {
		return current;
	}

	pointer operator->() const {
		return current;
	}

	reference operator*() const {
		return std::move(reinterpret_cast < value_type & >(*current));
	}

	map_move_iterator& operator++() {
		++current;
		return *this;
	}

	map_move_iterator& operator--() {
		--current;
		return *this;
	}

	map_move_iterator operator++(int) {
		map_move_iterator temp = *this;
		++current;
		return temp;
	}

	map_move_iterator operator--(int) {
		map_move_iterator temp = *this;
		--current;
		return temp;
	}

	bool operator== (const map_move_iterator<Iterator, KeyType, ValueType>& other) {
		return this->current == other.current;
	}

	bool operator!= (const map_move_iterator<Iterator, KeyType, ValueType>& other) {
		return ! ( *this == other );
	}

};

template<class T, class R>
class moveable_map_ref {
	ext::map < T, R > & theMap;
public:
	moveable_map_ref ( ext::map < T, R > & param ) : theMap ( param ) {}

	map_move_iterator < typename ext::map < T, R >::iterator, T, R > begin ( ) {
		return map_move_iterator < typename ext::map < T, R >::iterator, T, R > ( theMap.begin ( ) );
	}

	map_move_iterator < typename ext::map < T, R >::iterator, T, R > end ( ) {
		return map_move_iterator < typename ext::map < T, R >::iterator, T, R > ( theMap.end ( ) );
	}
};

template<class T, class R>
moveable_map_ref < T, R > make_moveable_map (ext::map < T, R > & map) {
	return moveable_map_ref < T, R > ( map );
}

template<class T, class R>
class moveable_map {
	ext::map<T, R> theMap;
public:
	moveable_map ( ext::map < T, R > && param) : theMap(std::move(param)) {}

	map_move_iterator<typename ext::map < T, R >::iterator, T, R > begin() {
		return map_move_iterator<typename ext::map < T, R >::iterator, T, R > (theMap.begin());
	}

	map_move_iterator<typename ext::map < T, R >::iterator, T, R > end() {
		return map_move_iterator<typename ext::map < T, R >::iterator, T, R > (theMap.end());
	}
};

template < class T, class R >
moveable_map < T, R > make_moveable_map ( ext::map < T, R > && map) {
	return moveable_map < T, R > ( std::move ( map ) );
}

} /* namespace ext */

#endif /* __MAP_HPP_ */
